#include <string.h>

#include "../../sr_module.h"
#include "../../dprint.h"
#include "../../mem/mem.h"
#include "../../parser/parse_from.h"
#include "../../parser/parse_uri.h"
#include "../../data_lump.h"
#include "../../mod_fix.h"
#include "../../ut.h"
#include "../../rw_locking.h"
#include "../../timer.h"
#include "../../db/db.h"
#include "new_mod.h"

#define USER_AGENT_HDR "User-Agent: "
#define USER_AGENT_HDR_LEN (sizeof(USER_AGENT_HDR)-1)

#define USER_AGENT_SUFFIX "\r\n"
#define USER_AGENT_SUFFIX_LEN 2

static int mod_init(void);
static int child_init(int);
static void destroy(void);
static int set_mapping(unsigned int type, void *val);
static int get_mapping(struct sip_msg *msg,char *value);
static int get_mapping0(struct sip_msg *msg,char *value);
static struct mi_root* mi_get_maps(struct mi_root* root, void *param);

struct ua_mapping *static_ua_map = NULL;
struct ua_mapping **db_ua_map = NULL;
static str db_url;
static str table_name=str_init("new_module");
static str value_col=str_init("value");
static str translated_col=str_init("translated");

static str mapping_value=str_init("$avp(map_value)");
static pv_spec_t map_value_spec;

static db_func_t db_funcs;
static db_con_t *db_con;

int enable_stats=1;
stat_var *translations_no = 0;
unsigned long get_translations_ps(void* foo);

void mapping_timer(unsigned int ticks,void *attr);
static rw_lock_t *ref_lock = NULL;
void new_module_process(int rank);

/* timer */
int timer_interval=10;

static cmd_export_t cmds[]={
	{"get_mapping",(cmd_function)get_mapping,1,fixup_pvar_null,
		0, REQUEST_ROUTE|ONREPLY_ROUTE},
	{"get_mapping",(cmd_function)get_mapping0,0,0,
		0, REQUEST_ROUTE|ONREPLY_ROUTE},
	{0,0,0,0,0,0}
};

static param_export_t params[]={
	{ "static_map", STR_PARAM|USE_FUNC_PARAM, (void *)&set_mapping},
	{ "db_url",     STR_PARAM, &db_url.s},
	{ "mapping_value", STR_PARAM, &mapping_value.s},
	{ "enable_stats", INT_PARAM, &enable_stats},
	{ "timer_interval", INT_PARAM, &timer_interval},
	{0,0,0}
};

static mi_export_t mi_cmds[] = {
	{ "get_maps","return all mappings",mi_get_maps,MI_NO_INPUT_FLAG,0,0},
	{0,0,0,0,0,0}
};

static stat_export_t mod_stats[] = {
	{"translations_no",         STAT_NO_RESET,  &translations_no},
	{"translations_per_sec",    STAT_IS_FUNC,   (stat_var**)get_translations_ps},
	{0,0,0}
};

static proc_export_t procs[] = {
	{"New module extra process",  0,  0, new_module_process, 1, 0},
	{0,0,0,0,0,0}
};

/** module exports */
struct module_exports exports= {
	"new_module",               /* module name */
	MODULE_VERSION,
	DEFAULT_DLFLAGS,            /* dlopen flags */
	cmds,                       /* exported functions */
	params,                     /* exported parameters */
	mod_stats,                  /* exported statistics */ 
	mi_cmds,                    /* exported MI functions */
	0,                          /* exported pseudo-variables */
	procs,                      /* extra processes */
	mod_init,                   /* module initialization function */
	(response_function) 0,      /* response handling function */
	(destroy_function)destroy,  /* destroy function */
	child_init                  /* per-child init function */
};

static int set_mapping(unsigned int type, void *val)
{
	char *value,*translated,*p;
	int value_len,translated_len,len;
	struct ua_mapping *new_map;

	value=(char *)val;
	len = strlen(value);

	p = memchr(value,MAPPING_DELIM,len);
	if (!p) {
		LM_ERR("Invalid mapping - delimiter not found\n");
		return -1;
	}

	value_len = p-value;
	translated=++p;

	translated_len = len - value_len-1;
	new_map = pkg_malloc(sizeof(struct ua_mapping));
	if (!new_map) {
		LM_ERR("No more pkg memory\n");
		return -1;
	}

	new_map->value.s = value;
	new_map->value.len = value_len;
	new_map->translated.s = translated;
	new_map->translated.len = translated_len;

	if (!static_ua_map)
		static_ua_map = new_map;
	else {
		new_map->next = static_ua_map;
		static_ua_map = new_map;
	}

	return 0;
}

static int mod_init(void)
{
	LM_INFO("Initializing module\n");
	if (db_url.s) {
		db_url.len = strlen(db_url.s);
		LM_INFO("We have db_url = %.*s\n",db_url.len,db_url.s);

		/* Find a database module */
		if (db_bind_mod(&db_url, &db_funcs) < 0){
			LM_ERR("Unable to bind to a database driver\n");
			return -1;
		}

		/* open a test connection */
		if ((db_con = db_funcs.init(&db_url)) == 0) {
			LM_ERR("cannot init connection to DB\n");
			return -1;
		}

		db_funcs.close(db_con);
		db_con = 0;
	}

	/* if statistics are disabled, prevent their registration to core */
	if (enable_stats==0)
		exports.stats = 0;

	mapping_value.len = strlen(mapping_value.s);
	LM_INFO("We'll try to store the mapping result in %.*s\n",
			mapping_value.len,mapping_value.s);
	if (!pv_parse_spec(&mapping_value,&map_value_spec)) {
		LM_ERR("invalid variable\n");
		return -1;
	}

	if ( register_timer( mapping_timer, 0, timer_interval)<0 ) {
		LM_ERR("failed to register timer \n");
		return -1;
	}

	db_ua_map = shm_malloc(sizeof(struct ua_mapping *));
	if (!db_ua_map) {
		LM_ERR("no more shm\n");
		return -1;
	}

	*db_ua_map = NULL;

	if ((ref_lock = lock_init_rw()) == NULL) {
		LM_ERR("failed to init lock\n");
		return -1;
	}

	return 0;
}

static int child_init(int rank)
{
	LM_INFO("Initializing child\n");
	if (db_url.s && rank == PROC_TIMER) {
		/* open a test connection */
		if ((db_con = db_funcs.init(&db_url)) == 0) {
			LM_ERR("cannot init connection to DB\n");
			return -1;
		}
		/* populate information on startup */
		mapping_timer(0,0);
	}

	return 0;
}

/*
 * destroy function
 */
static void destroy(void)
{
	struct ua_mapping *it,*aux;

	LM_INFO("Destroying module\n");

	it = static_ua_map;
	while (it) {
		aux = it;
		it=it->next;

		pkg_free(aux);
	}

	db_funcs.close(db_con);
}

static int get_mapping(struct sip_msg *msg,char *param)
{
	struct ua_mapping *it;
	pv_spec_p spec;
	pv_value_t spec_value;
	str *search_value;
	str replacement;
	struct to_body *pfrom;
	struct sip_uri puri;
	struct lump *l;
	char *s,*p;
	int len;

	spec = (pv_spec_p)param;

	/* print method and RURI */
	LM_DBG("Request method is [%.*s] (%d) \n",
			msg->first_line.u.request.method.len,msg->first_line.u.request.method.s,
			msg->first_line.u.request.method_value);
	LM_DBG("Request RURI is [%.*s] \n",
			GET_RURI(msg)->len, GET_RURI(msg)->s);
	/* print src and dst from IP level */
	LM_DBG("SRC info IP=[%s] port=%d proto=%d \n",
			ip_addr2a(&msg->rcv.src_ip), msg->rcv.src_port, msg->rcv.proto);
	LM_DBG("DST info IP=[%s] port=%d proto=%d \n",
			ip_addr2a(&msg->rcv.dst_ip), msg->rcv.dst_port, msg->rcv.proto);

	if (msg->flags & (1<<3))
		LM_DBG("FLAG 3 is set\n");
	else
		LM_DBG("FLAG 3 is not set\n");

	if ( parse_headers( msg, HDR_TO_F|HDR_FROM_F|HDR_USERAGENT_F, 0)!=0 ) {
		LM_ERR("failed to parse sip message\n");
		return -1;
	}
	if (msg->from==NULL && spec==NULL) {
		LM_ERR("no FROM hdr into message & no variable\n");
		return -1;
	}
	if (msg->to==NULL) {
		LM_ERR("no TO hdr into message\n");
		return -1;
	}

	if (spec) {
		if (pv_get_spec_value(msg,spec,&spec_value)!= 0) {
			LM_ERR("Cannot get spec value\n");
			return -1;
		}
		if (!(spec_value.flags & PV_VAL_STR)) {
			LM_ERR("Failure - value is not a string\n");
			return -1;
		}
		search_value = &spec_value.rs;
	} else {
		if ( parse_from_header( msg )!=0 ) {
			LM_ERR("bad from header [%.*s]\n",
					msg->from->body.len, msg->from->body.s);
			return -1;
		}
		pfrom = get_from(msg);
		LM_DBG("FROM: URI=[%.*s] DISPLAY=[%.*s]\n",
				pfrom->uri.len,pfrom->uri.s, pfrom->display.len,pfrom->display.s);
		if ( parse_uri( pfrom->uri.s, pfrom->uri.len, &puri)!=0 ) {
			LM_ERR("SIP URI in FROM hdr is not valid\n");
			return -1;
		}

		search_value = &puri.user;
	}

	LM_DBG("Searching for [%.*s]\n",search_value->len,search_value->s);

	/* changes into the message */
	if (msg->user_agent) {
		LM_DBG("USERAGENT: [%.*s]\n",
				msg->user_agent->body.len, msg->user_agent->body.s);
		l = del_lump( msg, msg->user_agent->name.s - msg->buf,
				msg->user_agent->len,
				HDR_USERAGENT_T);
		if (l==NULL) {
			LM_ERR("failed to add del lump\n");
			return -1;
		}
	} else {
		LM_DBG("USERAGENT hdr not found\n");
		l = anchor_lump( msg, msg->from->body.s + msg->from->body.len - msg->buf,
				0, HDR_USERAGENT_T );
		if (l==NULL) {
			LM_ERR("failed to create anchor lump\n");
			return -1;
		}
	}

	/* search through static mappings - have priority over DB */
	for (it=static_ua_map;it;it=it->next)
		if (it->value.len == search_value->len &&
				memcmp(it->value.s,search_value->s,it->value.len) == 0) {
			LM_INFO("Found mapping to %.*s\n",it->translated.len,it->translated.s);
			replacement = it->translated;
			break;
		}

	/* not found in static mappings, search in DB mappings*/
	if (!it) {
		lock_start_read( ref_lock );
		for (it=*db_ua_map;it;it=it->next)
			if (it->value.len == search_value->len &&
					memcmp(it->value.s,search_value->s,it->value.len) == 0) {
				LM_INFO("Found mapping to %.*s\n",it->translated.len,it->translated.s);
				replacement = it->translated;
				break;
			}
		lock_stop_read( ref_lock );
	}

	if (!it) {
		LM_ERR("no mappings found\n");
		return -1;
	}

	len = USER_AGENT_HDR_LEN + replacement.len + USER_AGENT_SUFFIX_LEN;
	s = pkg_malloc(len);
	if (s==NULL) {
		LM_ERR("no more pkg memory\n");
		return -1;
	}
	p = s;
	memcpy(p,USER_AGENT_HDR,USER_AGENT_HDR_LEN);
	p+=USER_AGENT_HDR_LEN;
	memcpy(p,replacement.s,replacement.len);
	p+=replacement.len;
	memcpy(p,USER_AGENT_SUFFIX,USER_AGENT_SUFFIX_LEN);
	l = insert_new_lump_after( l, s, len , HDR_USERAGENT_T);
	if (l==NULL) {
		LM_ERR("failed to insert new lump\n");
		return -1;
	}

	/* setting output variable */
	spec_value.rs = replacement;
	spec_value.flags = PV_VAL_STR;
	if (pv_set_value(msg,&map_value_spec,0,&spec_value) != 0)
		LM_ERR("Failed to set the output variable\n");

	if_update_stat( enable_stats, translations_no , 1);

	return 1;
}

static int get_mapping0(struct sip_msg *msg,char *param)
{
	return get_mapping(msg,NULL);
}

unsigned long get_translations_ps(void* foo)
{
	int trans_no;
	int uptime;

	trans_no = get_stat_val(translations_no);
	uptime = get_ticks();

	return trans_no/uptime;
}

void mapping_timer(unsigned int ticks,void *attr)
{
	db_key_t query_cols[] = { &value_col,&translated_col };
	db_res_t * res;
	db_val_t * values;
	db_row_t * rows;
	str value,translated;
	int nr_rows,i,size;
	struct ua_mapping *new_map,*it,*old_list,*aux;
	struct ua_mapping *new_list=NULL;

	LM_INFO("timer running at %u tick\n",ticks);
	/* replacement */
	db_funcs.use_table(db_con, &table_name);

	/* select value,translated from new_module */
	if (db_funcs.query(db_con,0,0,0,query_cols,0,2,0,&res) != 0) {
		LM_ERR("Failure to issue query\n");
		return;
	}

	nr_rows = RES_ROW_N(res);
	rows = RES_ROWS(res);

	for (i=0;i<nr_rows;i++) {
		values = ROW_VALUES(rows+i);

		if (VAL_NULL(values) ||
				(VAL_TYPE(values) != DB_STR && VAL_TYPE(values) != DB_STRING)) {
			LM_ERR("Invalid value returned\n");
			goto end;
		}

		if (VAL_NULL(values+1) ||
				(VAL_TYPE(values+1) != DB_STR && VAL_TYPE(values+1) != DB_STRING)) {
			LM_ERR("Invalid translated returned\n");
			goto end;
		}

		if (VAL_TYPE(values) == DB_STR) {
			value = VAL_STR(values);
		} else {
			value.s = (char*)VAL_STRING(values);
			value.len = strlen(value.s);
		}
		if (value.len <= 0) {
			LM_ERR("Empty translation value\n");
			goto end;
		}

		if (VAL_TYPE(values+1) == DB_STR) {
			translated = VAL_STR(values+1);
		} else {
			translated.s = (char*)VAL_STRING(values+1);
			translated.len = strlen(translated.s);
		}
		if (translated.len <= 0) {
			LM_ERR("Empty translation result\n");
			goto end;
		}

		size = sizeof(struct ua_mapping)+value.len+translated.len;
		new_map = shm_malloc(size);
		if (!new_map) {
			LM_ERR("no more shm\n");
			goto end;
		}

		memset(new_map,0,size);

		new_map->value.len = value.len;
		new_map->value.s=(char *)(new_map+1);
		memcpy(new_map->value.s,value.s,value.len);
		new_map->translated.len = translated.len;
		new_map->translated.s = (char *)(new_map+1) + value.len;
		memcpy(new_map->translated.s,translated.s,translated.len);

		LM_INFO("Loaded translation [%.*s]->[%.*s]\n",value.len,value.s,
				translated.len,translated.s);
		if (new_list) {
			new_list->next = new_map;
			new_list = new_map;
		} else
			new_list = new_map;
	}

	lock_start_write( ref_lock );
	old_list = *db_ua_map;
	*db_ua_map = new_list;
	lock_stop_write( ref_lock );

	it = old_list;
	while (it) {
		aux = it;
		it=it->next;

		shm_free(aux);
	}

end:
	db_funcs.free_result(db_con, res);
}

static struct mi_root* mi_get_maps(struct mi_root* root, void *param)
{
	struct mi_root* rpl_tree= NULL;
	struct mi_node* node= NULL,*rpl=NULL;
	struct mi_attr* attr= NULL;
	struct ua_mapping *it;

	rpl_tree = init_mi_tree( 200, MI_SSTR(MI_OK));
	if (!rpl_tree) {
		LM_ERR("failed initializing MI rpl tree\n");
		goto error;
	}

	rpl = &rpl_tree->node;

	for (it=static_ua_map;it;it=it->next) {
		node=add_mi_node_child(rpl,0,"Map",3,0,0);
		if (!node)
			goto error;
		attr=add_mi_attr(node,0,"value",5,it->value.s,it->value.len);
		if (!attr)
			goto error;
		attr=add_mi_attr(node,0,"translated",10,it->translated.s,it->translated.len);
		if (!attr)
			goto error;
	}

	return rpl_tree;

error:
	/* trash everything that was built so far */
	if (rpl_tree) free_mi_tree(rpl_tree);
	return NULL;
}


void new_module_process(int rank)
{
	LM_INFO("Started new process with rank %d\n", rank);

	/* do asynchronous work */
}
